@@ -684,29 +684,25 @@ typedef struct {
 #define IO_QUEUE_EXTSTORE 1
 #define IO_QUEUE_PROXY 2
 
+#define IO_PENDING_TYPE_NONE 0
+#define IO_PENDING_TYPE_BASIC 1
+#define IO_PENDING_TYPE_EXTSTORE 2
+#define IO_PENDING_TYPE_PROXY 3
+
 typedef STAILQ_HEAD(iop_head_s, _io_pending_t) iop_head_t;
 typedef struct _io_pending_t io_pending_t;
 typedef struct io_queue_s io_queue_t;
 typedef void (*io_queue_stack_cb)(io_queue_t *q);
 typedef void (*io_queue_cb)(io_pending_t *pending);
-// This structure used to be passed between threads, but is now owned entirely
-// by the worker threads.
 // IO pending objects are created and stacked into this structure. They are
 // then sent off to remote threads.
-// The objects are returned one at a time to the worker threads, and this
-// structure is then consulted to see when to resume the worker.
-struct io_queue_s {
-    void *ctx; // duplicated from io_queue_cb_t
-    void *stack_ctx; // module-specific context to be batch-submitted
-    int count; // ios to process before returning. only accessed by queue processor once submitted
-    int type; // duplicated from io_queue_cb_t
-};
-
-typedef struct io_queue_cb_s {
+// The objects are returned one at a time to the worker threads.
+typedef struct io_queue_s {
     void *ctx; // untouched ptr for specific context
+    iop_head_t stack;
     io_queue_stack_cb submit_cb; // callback given a full stack of pending IO's at once.
     int type;
-} io_queue_cb_t;
+} io_queue_t;
 
 struct thread_notify {
     struct event notify_event;  /* listen event for notify pipe or eventfd */
@@ -728,8 +724,9 @@ typedef struct {
     iop_head_t ion_head;        /* queue for IO object return */
     int cur_sfd;                /* client fd for logging commands */
     int thread_baseid;          /* which "number" thread this is for data offsets */
+    int conns_tosubmit;         /* number of conns which have put data in io_queue */
     struct thread_stats stats;  /* Stats generated by this thread */
-    io_queue_cb_t io_queues[IO_QUEUE_COUNT];
+    io_queue_t io_queues[IO_QUEUE_COUNT];
     struct conn_queue *ev_queue; /* Worker/conn event queue */
     cache_t *rbuf_cache;        /* static-sized read buffers */
     mc_resp_bundle *open_bundle;
@@ -787,6 +784,7 @@ typedef struct _mc_resp {
      * to asynchronously kill an object that was queued to write
      */
     bool skip;
+    bool suspended; // waiting for response from subsystem
     bool free; // double free detection.
 #ifdef PROXY
     bool proxy_res; // we're handling a proxied response buffer.
@@ -814,7 +812,9 @@ struct _mc_resp_bundle {
 typedef struct conn conn;
 
 struct _io_pending_t {
-    int io_queue_type; // matches one of IO_QUEUE_*
+    uint8_t io_queue_type; // which queue this entered on
+    uint8_t io_sub_type; // sub-type. FIXME: can optimize out later.
+    uint8_t payload;
     LIBEVENT_THREAD *thread;
     conn *c;
     mc_resp *resp; // associated response object
@@ -870,8 +870,7 @@ struct conn {
     /* data for the swallow state */
     int    sbytes;    /* how many bytes to swallow */
 
-    int io_queues_submitted; /* see notes on io_queue_t */
-    io_queue_t io_queues[IO_QUEUE_COUNT]; /* set of deferred IO queues. */
+    int resps_suspended; /* see notes on io_queue_cb_t */
 #ifdef PROXY
     void *proxy_rctx; /* pointer to active request context */
 #endif
@@ -937,10 +936,24 @@ enum delta_result_type do_add_delta(LIBEVENT_THREAD *t, const char *key,
                                     item **it_ret);
 enum store_item_type do_store_item(item *item, int comm, LIBEVENT_THREAD *t, const uint32_t hv, int *nbytes, uint64_t *cas, const uint64_t cas_in, bool cas_stale);
 void thread_io_queue_add(LIBEVENT_THREAD *t, int type, void *ctx, io_queue_stack_cb cb);
-void conn_io_queue_setup(conn *c);
-io_queue_t *conn_io_queue_get(conn *c, int type);
-io_queue_cb_t *thread_io_queue_get(LIBEVENT_THREAD *t, int type);
+io_queue_t *thread_io_queue_get(LIBEVENT_THREAD *t, int type);
+void thread_io_queue_submit(LIBEVENT_THREAD *t);
 void conn_io_queue_return(io_pending_t *io);
+#define conn_resp_suspend(c, resp) \
+    do { \
+        resp->suspended = true; \
+        c->resps_suspended++; \
+    } while (0)
+#define conn_resp_unsuspend(c, resp) \
+    do { \
+        resp->suspended = false; \
+        c->resps_suspended--; \
+        assert(c->resps_suspended >= 0); \
+        if (c->resps_suspended == 0) { \
+            conn_worker_readd(c); \
+        } \
+    } while (0)
+
 conn *conn_new(const int sfd, const enum conn_states init_state, const int event_flags, const int read_buffer_size,
     enum network_transport transport, struct event_base *base, void *ssl, uint64_t conntag, enum protocol bproto);
 
